from __future__ import annotations

from collections.abc import Iterable

from pants.backend.python.util_rules.interpreter_constraints import InterpreterConstraints
from pants.core.goals.resolves import ExportableTool
from pants.core.util_rules.config_files import ConfigFilesRequest
from pants.core.util_rules.external_tool import TemplatedExternalTool
from pants.engine.platform import Platform
from pants.engine.rules import Rule, collect_rules
from pants.engine.unions import UnionRule
from pants.option.option_types import (
    ArgsListOption,
    BoolOption,
    FileOption,
    SkipOption,
    StrListOption,
)
from pants.util.strutil import softwrap

from pants_ty.known_versions import DEFAULT_URL_PLATFORM_MAPPING, DEFAULT_URL_TEMPLATE


class Ty(TemplatedExternalTool):
    options_scope = "ty"
    name = "Ty"
    help = softwrap(
        """
        The Ty Python type checker (https://docs.astral.sh/ty/).
        """
    )

    default_version = "0.0.28"
    default_known_versions = [
        "0.0.28|linux_arm64|79a3fdc5f06c1c8638808cc6b0fd8dd79cbffa252e0ff8d65383962baf1dd979|9979904",
        "0.0.28|linux_x86_64|0708bbd1786df4124aa16b6b3301322cc2a74637d74a72d04a4cce87d0c88d61|10827842",
        "0.0.28|macos_arm64|6bbd117bf1b0eb13f32e095eb19b9252cf427decc7b97eb1b3ed7cec127d42c9|9521516",
        "0.0.28|macos_x86_64|ea08ae80a7a9cc92ac5de050f26a9e3cb6253bdd39dbc9131be747f3e3814382|10168678",
    ]
    default_url_template = DEFAULT_URL_TEMPLATE
    default_url_platform_mapping = DEFAULT_URL_PLATFORM_MAPPING

    skip = SkipOption("check")
    args = ArgsListOption(example="--error-on-warning --output-format=concise")
    config = FileOption(
        default=None,
        advanced=True,
        help=softwrap(
            """
            Path to a `ty.toml` file to use for configuration.

            Setting this option disables `[ty].config_discovery`.
            """
        ),
    )
    config_discovery = BoolOption(
        default=True,
        advanced=True,
        help=softwrap(
            """
            If true, Pants will include relevant Ty config files during runs
            (`ty.toml` and `pyproject.toml` containing `[tool.ty]`).
            """
        ),
    )
    _interpreter_constraints = StrListOption(
        advanced=True,
        default=["CPython>=3.8,<3.15"],
        help="Python interpreter constraints for Ty.",
    )

    def generate_exe(self, plat: Platform) -> str:
        return f"ty-{self.default_url_platform_mapping[plat.value]}/ty"

    @property
    def interpreter_constraints(self) -> InterpreterConstraints:
        return InterpreterConstraints(self._interpreter_constraints)

    def config_request(self) -> ConfigFilesRequest:
        return ConfigFilesRequest(
            specified=self.config,
            specified_option_name=f"{self.options_scope}.config",
            discovery=self.config_discovery,
            check_existence=["ty.toml"],
            check_content={"pyproject.toml": b"[tool.ty"},
        )


def rules() -> Iterable[Rule | UnionRule]:
    return (
        *collect_rules(),
        *Ty.rules(),
        UnionRule(ExportableTool, Ty),
    )



from __future__ import annotations

from collections.abc import Iterable

from pants.engine.rules import Rule
from pants.engine.unions import UnionRule

from pants_ty import rules as ty_rules
from pants_ty import skip_field, subsystem


def rules() -> Iterable[Rule | UnionRule]:
    return (*subsystem.rules(), *skip_field.rules(), *ty_rules.rules())



from __future__ import annotations

import argparse
import hashlib
from collections.abc import Mapping, Sequence
from urllib.request import urlopen

DEFAULT_URL_TEMPLATE = (
    "https://github.com/astral-sh/ty/releases/download/{version}/ty-{platform}.tar.gz"
)
DEFAULT_URL_PLATFORM_MAPPING = {
    "linux_arm64": "aarch64-unknown-linux-musl",
    "linux_x86_64": "x86_64-unknown-linux-musl",
    "macos_arm64": "aarch64-apple-darwin",
    "macos_x86_64": "x86_64-apple-darwin",
}


def _normalize_platforms(
    platforms: Sequence[str] | None,
    *,
    platform_mapping: Mapping[str, str],
) -> tuple[str, ...]:
    if not platforms:
        return tuple(platform_mapping.keys())

    unknown_platforms = tuple(platform for platform in platforms if platform not in platform_mapping)
    if unknown_platforms:
        expected = ", ".join(sorted(platform_mapping))
        unknown = ", ".join(unknown_platforms)
        raise ValueError(f"Unknown Pants platform(s): {unknown}. Expected one of: {expected}")

    return tuple(platforms)


def _download_metadata(url: str) -> tuple[str, int]:
    digest = hashlib.sha256()
    length = 0

    with urlopen(url) as response:
        while chunk := response.read(64 * 1024):
            digest.update(chunk)
            length += len(chunk)

    return digest.hexdigest(), length


def generate_known_versions(
    version: str,
    *,
    url_template: str = DEFAULT_URL_TEMPLATE,
    platform_mapping: Mapping[str, str] = DEFAULT_URL_PLATFORM_MAPPING,
    platforms: Sequence[str] | None = None,
) -> tuple[str, ...]:
    entries: list[str] = []

    for pants_platform in _normalize_platforms(platforms, platform_mapping=platform_mapping):
        upstream_platform = platform_mapping[pants_platform]
        url = url_template.format(version=version, platform=upstream_platform)
        sha256, length = _download_metadata(url)
        entries.append(f"{version}|{pants_platform}|{sha256}|{length}")

    return tuple(entries)


def format_known_versions_block(version: str, entries: Sequence[str]) -> str:
    lines = [
        "[ty]",
        f'version = "{version}"',
        "known_versions = [",
        *(f'  "{entry}",' for entry in entries),
        "]",
    ]
    return "\n".join(lines)


def create_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Generate Pants [ty].known_versions entries for a ty release.",
    )
    parser.add_argument("version", help="ty version to inspect, for example 0.0.27")
    parser.add_argument(
        "--platform",
        dest="platforms",
        action="append",
        choices=tuple(DEFAULT_URL_PLATFORM_MAPPING),
        help="Limit output to one Pants platform. May be passed multiple times.",
    )
    parser.add_argument(
        "--url-template",
        default=DEFAULT_URL_TEMPLATE,
        help=(
            "Download URL template. Use {version} and {platform}. "
            "Defaults to the official Astral GitHub release archive pattern."
        ),
    )
    parser.add_argument(
        "--entries-only",
        action="store_true",
        help="Print only the known_versions entries instead of a full [ty] block.",
    )
    return parser


def main(argv: Sequence[str] | None = None) -> int:
    parser = create_argument_parser()
    args = parser.parse_args(argv)
    version = args.version.removeprefix("v")
    entries = generate_known_versions(
        version,
        url_template=args.url_template,
        platforms=args.platforms,
    )

    if args.entries_only:
        print("\n".join(f'"{entry}",' for entry in entries))
    else:
        print(format_known_versions_block(version, entries))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import os
from collections.abc import Iterable, Sequence
from dataclasses import dataclass

from pants.backend.python.subsystems.setup import PythonSetup
from pants.backend.python.target_types import (
    InterpreterConstraintsField,
    PythonResolveField,
    PythonSourceField,
)
from pants.backend.python.util_rules import pex_from_targets
from pants.backend.python.util_rules.interpreter_constraints import InterpreterConstraints
from pants.backend.python.util_rules.partition import (
    _partition_by_interpreter_constraints_and_resolve,
)
from pants.backend.python.util_rules.pex import (
    PexRequest,
    VenvPexProcess,
    VenvPexRequest,
    create_pex,
    create_venv_pex,
)
from pants.backend.python.util_rules.pex_environment import PexEnvironment
from pants.backend.python.util_rules.pex_from_targets import RequirementsPexRequest
from pants.backend.python.util_rules.python_sources import (
    PythonSourceFilesRequest,
    prepare_python_sources,
)
from pants.core.goals.check import CheckRequest, CheckResult, CheckResults
from pants.core.util_rules import config_files
from pants.core.util_rules.config_files import find_config_file
from pants.core.util_rules.external_tool import download_external_tool
from pants.core.util_rules.source_files import SourceFilesRequest, determine_source_files
from pants.engine.collection import Collection
from pants.engine.fs import MergeDigests
from pants.engine.internals.graph import resolve_coarsened_targets
from pants.engine.intrinsics import execute_process, merge_digests
from pants.engine.platform import Platform
from pants.engine.process import Process, ProcessCacheScope, execute_process_or_raise
from pants.engine.rules import Rule, collect_rules, concurrently, implicitly, rule
from pants.engine.target import (
    CoarsenedTargets,
    CoarsenedTargetsRequest,
    FieldSet,
    Target,
)
from pants.engine.unions import UnionRule
from pants.util.logging import LogLevel
from pants.util.ordered_set import FrozenOrderedSet, OrderedSet
from pants.util.strutil import pluralize

from pants_ty.skip_field import SkipTyField
from pants_ty.subsystem import Ty


def _python_version_args(
    *, python_version: str | None, user_args: Sequence[str]
) -> tuple[str, ...]:
    if python_version is None:
        return ()
    if any(
        arg.startswith("--python-version=") or arg.startswith("--target-version=")
        for arg in user_args
    ):
        return ()
    return (f"--python-version={python_version}",)


def _extra_search_path_args(source_roots: Sequence[str]) -> tuple[str, ...]:
    unique_source_roots = tuple(dict.fromkeys(source_roots))
    return tuple(f"--extra-search-path={source_root}" for source_root in unique_source_roots)


def _batch_input_paths(
    paths: Sequence[str], *, max_paths_per_batch: int
) -> tuple[tuple[str, ...], ...]:
    if max_paths_per_batch <= 0:
        raise ValueError("max_paths_per_batch must be positive")
    if not paths:
        return ()
    return tuple(
        tuple(paths[index : index + max_paths_per_batch])
        for index in range(0, len(paths), max_paths_per_batch)
    )


@dataclass(frozen=True)
class TyFieldSet(FieldSet):
    required_fields = (PythonSourceField,)

    sources: PythonSourceField
    resolve: PythonResolveField
    interpreter_constraints: InterpreterConstraintsField

    @classmethod
    def opt_out(cls, tgt: Target) -> bool:
        return tgt.get(SkipTyField).value


class TyRequest(CheckRequest):
    field_set_type = TyFieldSet
    tool_name = Ty.options_scope


@dataclass(frozen=True)
class TyPartition:
    field_sets: FrozenOrderedSet[TyFieldSet]
    root_targets: CoarsenedTargets
    resolve_description: str | None
    interpreter_constraints: InterpreterConstraints

    def description(self) -> str:
        ics = str(sorted(str(c) for c in self.interpreter_constraints))
        return f"{self.resolve_description}, {ics}" if self.resolve_description else ics


class TyPartitions(Collection[TyPartition]):
    pass


@rule(
    desc="Determine if it is necessary to partition Ty's input",
    level=LogLevel.DEBUG,
)
async def ty_determine_partitions(
    request: TyRequest,
    ty: Ty,
    python_setup: PythonSetup,
) -> TyPartitions:
    resolve_and_interpreter_constraints_to_field_sets = (
        _partition_by_interpreter_constraints_and_resolve(request.field_sets, python_setup)
    )

    coarsened_targets = await resolve_coarsened_targets(
        CoarsenedTargetsRequest(field_set.address for field_set in request.field_sets),
        **implicitly(),
    )
    coarsened_targets_by_address = coarsened_targets.by_address()

    return TyPartitions(
        TyPartition(
            FrozenOrderedSet(field_sets),
            CoarsenedTargets(
                OrderedSet(
                    coarsened_targets_by_address[field_set.address] for field_set in field_sets
                )
            ),
            resolve if len(python_setup.resolves) > 1 else None,
            interpreter_constraints or ty.interpreter_constraints,
        )
        for (resolve, interpreter_constraints), field_sets in sorted(
            resolve_and_interpreter_constraints_to_field_sets.items()
        )
    )


@rule(desc="Typecheck one Ty partition", level=LogLevel.DEBUG)
async def ty_typecheck_partition(
    partition: TyPartition,
    ty: Ty,
    platform: Platform,
    pex_environment: PexEnvironment,
    python_setup: PythonSetup,
) -> CheckResult:
    (
        ty_tool,
        config_files_snapshot,
        root_sources,
        transitive_sources,
        requirements_pex,
    ) = await concurrently(
        download_external_tool(ty.get_request(platform)),
        find_config_file(ty.config_request()),
        determine_source_files(SourceFilesRequest(fs.sources for fs in partition.field_sets)),
        prepare_python_sources(
            PythonSourceFilesRequest(partition.root_targets.closure()), **implicitly()
        ),
        create_pex(
            **implicitly(
                RequirementsPexRequest(
                    (fs.address for fs in partition.field_sets),
                    hardcoded_interpreter_constraints=partition.interpreter_constraints,
                )
            )
        ),
    )

    complete_pex_env = pex_environment.in_workspace()
    requirements_venv_pex = await create_venv_pex(
        VenvPexRequest(
            PexRequest(
                output_filename="requirements_venv.pex",
                internal_only=True,
                pex_path=[requirements_pex],
                interpreter_constraints=partition.interpreter_constraints,
            ),
            complete_pex_env,
        ),
        **implicitly(),
    )

    _ = await execute_process_or_raise(
        **implicitly(
            VenvPexProcess(
                requirements_venv_pex,
                description="Force Ty requirements venv to materialize",
                argv=["-c", "''"],
                cache_scope=ProcessCacheScope.PER_SESSION,
            )
        )
    )

    input_digest = await merge_digests(
        MergeDigests(
            (
                transitive_sources.source_files.snapshot.digest,
                config_files_snapshot.snapshot.digest,
                requirements_venv_pex.digest,
            )
        )
    )

    immutable_input_key = "__ty_tool"
    exe_path = os.path.join(immutable_input_key, ty_tool.exe)
    requirements_venv_path = os.path.join(
        complete_pex_env.pex_root, requirements_venv_pex.venv_rel_dir
    )

    python_version = partition.interpreter_constraints.minimum_python_version(
        python_setup.interpreter_versions_universe
    )
    config_args = (f"--config-file={ty.config}",) if ty.config else ()
    common_args = (
        "check",
        *config_args,
        f"--python={requirements_venv_path}",
        *_python_version_args(python_version=python_version, user_args=ty.args),
        *_extra_search_path_args(transitive_sources.source_roots),
        *ty.args,
    )
    path_batches = _batch_input_paths(tuple(root_sources.snapshot.files), max_paths_per_batch=512)

    if not path_batches:
        return CheckResult(0, "", "", partition_description=partition.description())

    batch_results = await concurrently(
        execute_process(
            Process(
                argv=(exe_path, *common_args, *path_batch),
                input_digest=input_digest,
                immutable_input_digests={immutable_input_key: ty_tool.digest},
                description=f"Run Ty on {pluralize(len(path_batch), 'file')}.",
                level=LogLevel.DEBUG,
            ),
            **implicitly(),
        )
        for path_batch in path_batches
    )

    return CheckResult(
        exit_code=max(result.exit_code for result in batch_results),
        stdout="".join(result.stdout.decode() for result in batch_results),
        stderr="".join(result.stderr.decode() for result in batch_results),
        partition_description=partition.description(),
    )


@rule(desc="Typecheck using Ty", level=LogLevel.DEBUG)
async def ty_typecheck(
    request: TyRequest,
    ty: Ty,
) -> CheckResults:
    if ty.skip:
        return CheckResults([], checker_name=request.tool_name)

    partitions = await ty_determine_partitions(request, **implicitly())
    partitioned_results = await concurrently(
        ty_typecheck_partition(partition, **implicitly()) for partition in partitions
    )
    return CheckResults(partitioned_results, checker_name=request.tool_name)


def rules() -> Iterable[Rule | UnionRule]:
    return (
        *collect_rules(),
        *config_files.rules(),
        *pex_from_targets.rules(),
        UnionRule(CheckRequest, TyRequest),
    )



from __future__ import annotations

from collections.abc import Iterable

from pants.backend.python.target_types import (
    PythonSourceTarget,
    PythonSourcesGeneratorTarget,
    PythonTestTarget,
    PythonTestsGeneratorTarget,
    PythonTestUtilsGeneratorTarget,
)
from pants.engine.rules import Rule
from pants.engine.target import BoolField
from pants.engine.unions import UnionRule


class SkipTyField(BoolField):
    alias = "skip_ty"
    default = False
    help = "If true, don't run Ty on this target's code."


def rules() -> Iterable[Rule | UnionRule]:
    return (
        PythonSourcesGeneratorTarget.register_plugin_field(SkipTyField),
        PythonSourceTarget.register_plugin_field(SkipTyField),
        PythonTestsGeneratorTarget.register_plugin_field(SkipTyField),
        PythonTestTarget.register_plugin_field(SkipTyField),
        PythonTestUtilsGeneratorTarget.register_plugin_field(SkipTyField),
    )
